@@ -1,7 +1,7 @@
 
 /*
     xskat - a card game for 1 to 3 players.
-    Copyright (C) 1996  Gunter Gerhardt
+    Copyright (C) 1997  Gunter Gerhardt
 
     This program is free software; you can redistribute it freely.
     Use it at your own risk; there is NO WARRANTY.
@@ -10,5 +10,8 @@
 #define static
 #include "icon.xbm"
 #include "bwcards.xbm"
+#include "bwcardl.xbm"
 #include "colcards.xbm"
-unsigned char colcards_pixs[colcards_width*colcards_height/4];
+#include "colcardl.xbm"
+char colcards_pixs[colcards_width*colcards_height/4];
+char colcardl_pixs[colcardl_width*colcardl_height/4];
