@@ -1,12 +1,15 @@
 
 /*
     xskat - a card game for 1 to 3 players.
-    Copyright (C) 1996  Gunter Gerhardt
+    Copyright (C) 1998  Gunter Gerhardt
 
     This program is free software; you can redistribute it freely.
     Use it at your own risk; there is NO WARRANTY.
 */
 
+#ifndef SKAT_H
+#define SKAT_H
+
 #undef EXTERN
 #ifdef SKAT_C
 #define EXTERN
@@ -17,8 +20,21 @@
 #define AS 0
 #define ZEHN 1
 #define KOENIG 2
+#define DAME 3
 #define BUBE 4
 #define NEUN 5
+#define ACHT 6
+#define SIEBEN 7
+
+#define BOCK_BEI_60         1
+#define BOCK_BEI_GRANDHAND  2
+#define BOCK_BEI_KONTRA     4
+#define BOCK_BEI_RE         8
+#define BOCK_BEI_NNN       16
+#define BOCK_BEI_N00       32
+#define BOCK_BEI_72        64
+#define BOCK_BEI_96       128
+#define BOCK_BEI_LAST     128
 
 enum {
  GEBEN,
@@ -27,6 +43,7 @@ enum {
  DRUECKEN,
  ANSAGEN,
  SPIELEN,
+ NIMMSTICH,
  RESULT
 };
 
@@ -79,28 +96,46 @@ EXTERN int sortw[]
 #endif
 ;
 
-EXTERN int numsp;
+EXTERN int numsp,numgames;
+EXTERN int strateg[3],stgset[3];
 EXTERN int maxrw[3],sort1[3],sort2[3],alternate[3],splfirst[3];
+EXTERN int downupset[3],altseqset[3],tdelayset[3],alistset[3];
+EXTERN int mbutton[3],mbuttonset[3];
+EXTERN int useoptfile[3];
 EXTERN int lastmsaho[3],protsort[3],hatnfb[3][5];
-EXTERN int splsum[3],sum[3],asplsum[3],asum[3],alist[3];
+EXTERN int splsum[3][3],sum[3][3],prevsum[3][3],alist[3];
+EXTERN int nimmstich[3][2],playramsch,ramschset;
+EXTERN int playsramsch,sramschset,sramschstufe;
+EXTERN int playkontra,kontraset,kontrastufe;
+EXTERN int playbock,bockset,bockevents,bockeventsset;
+EXTERN int bockspiele,ramschspiele,bockinc;
+EXTERN int resumebock,resumebockset;
+EXTERN int sagte18[3];
 EXTERN int cards[32],gespcd[32];
 EXTERN int quit,phase,geber,hoerer,sager,spieler;
 EXTERN int saho,reizp,gedr,vmh,stich,ausspl;
 EXTERN int possi[10],possc,stcd[3];
 EXTERN int trumpf,handsp,stsum,spcards[12];
-EXTERN int spgew,spwert,schwz,nullv;
+EXTERN int spgew,spwert,schwz,nullv,nspwert;
 EXTERN int schnang,schwang,ouveang;
-EXTERN int splist[360],splstp,splres;
-EXTERN int gespfb[4],high[5];
+EXTERN struct { int s,r,d,e,g; } splist[360];
+EXTERN int splstp,splres;
+EXTERN int gespfb[4],high[5],shigh[5];
 EXTERN int inhand[4][8];
 EXTERN int playcd,drkcd;
 EXTERN int mes1,mes2,mes3;
-EXTERN int predef,logging;
-EXTERN long seed[2],savseed,gamenr;
-EXTERN char *list_file,*game_file,*prot_file;
+EXTERN int butternok,sptruempfe;
+EXTERN int predef,logging,unformatted;
+EXTERN long seed[2],savseed,gamenr,rotateby;
+EXTERN char *list_file,*game_file,*prot_file,*opt_file;
+EXTERN int wieder,vorhandwn;
+EXTERN int karobubeanz;
 EXTERN struct
 {
-  int stichgem,spieler,trumpf,gereizt,gewonn,augen,ehsso,predef;
-  int stiche[10][3],anspiel[10],gemacht[10],skat[2][2];
+  int stichgem,spieler,trumpf,gereizt,gewonn,augen,ehsso,predef,sramsch;
+  int stiche[10][3],anspiel[10],gemacht[10],skat[4][2],verdopp[3];
   long savseed,gamenr;
+  int rotateby;
 } prot1,prot2;
+
+#endif /* SKAT_H */
